module modeling_mod

  implicit none
  
  real, private :: ds, dt, dt2, c2(0:1), c4(0:2), c6(0:3), c8(0:4), c10(0:5), c12(0:6), c14(0:7), c16(0:8)
  real, allocatable :: vel2(:,:), temp(:,:)
  real, parameter :: pi = 3.14159265359
  integer, private :: nx, nz

  contains 

  subroutine init_c16()
    ! Calculate coefficients of 16th order finite difference
    	  
    c16(1) = 1.
    c16(2) = -7./40.
    c16(3) = 7./165.
    c16(4) = -7./704.
    c16(5) = 7./3575.
    c16(6) = -1./3432.
    c16(7) = 1./35035.
    c16(8) = -1./732160.
    c16(0) = -2*sum(c16(1:8))
    c16 = 16./9.*c16
	  
  end subroutine

  subroutine init_c14()
    ! Calculate coefficients of 14th order finite difference

    c14(1) = 1.
    c14(2) = -1./6.
    c14(3) = 1./27.
    c14(4) = -1./132.
    c14(5) = 1./825.
    c14(6) = -1./7722.
    c14(7) = -1./147147.
    c14(0) = -2*sum(c14(1:7))
    c14 = 7./4.*c14
	  
  end subroutine

  subroutine init_c12()
    ! Calculate coefficients of 12th order finite difference

    c12(1) = 1.
    c12(2) = -5./32.
    c12(3) = 5./162.
    c12(4) = -1./192.
    c12(5) = 1./1650.
    c12(6) = -1./28512.
    c12(0) = -2*sum(c12(1:6))
    c12 = 12./7.*c12
	  
  end subroutine

  subroutine init_c10()
    ! Calculate coefficients of 10th order finite difference

    c10(1) = 1.
    c10(2) = -1./7.
    c10(3) = 1./42.
    c10(4) = -1./336.
    c10(5) = 1./5250.
    c10(0) = -2*sum(c10(1:5))
    c10 = 5./3.*c10
	  
  end subroutine

  subroutine init_c8()
    ! Calculate coefficients of 8th order finite difference

    c8(1) = 1.
    c8(2) = -1./8.
    c8(3) = 1./63.
    c8(4) = -1./896.
    c8(0) = -2*sum(c8(1:4))
    c8 = 8./5.*c8
	  
  end subroutine

  subroutine init_myc8()
    ! Calculate coefficients of 8th order finite difference

    c8(1) = 1.755863
    c8(2) = -0.290179
    c8(3) = 0.056295
    c8(4) = -0.0063626133
    c8(0) = -2*sum(c8(1:4))
	  
  end subroutine

  subroutine init_paperc8()
    ! Calculate coefficients of 8th order finite difference

    c8(1) = 1.70507669
    c8(2) = -0.25861812
    c8(3) = 0.04577745
    c8(4) = -0.0052363
    c8(0) = -2*sum(c8(1:4))
	  
  end subroutine

  subroutine init_c6()
    ! Calculate coefficients of 6th order finite difference

    c6(1) = 1.
    c6(2) = -1./10.
    c6(3) = 1./135.
    c6(0) = -2*sum(c6(1:3))
    c6 = 3./2.*c6
	  
  end subroutine

  subroutine init_c4()
    ! Calculate coefficients of 4th order finite difference

    c4(1) = 1.
    c4(2) = -1./16.
    c4(0) = -2*sum(c4(1:2))
    c4 = 4./3.*c4
	  
  end subroutine

  subroutine init_myc4()
    ! Calculate coefficients of 4th order finite difference

    c4(1) = 1.3662928
    c4(2) = -0.0915732
    c4(0) = -2*sum(c4(1:2))
	  
  end subroutine

  subroutine init_paperc4()
    ! Calculate coefficients of 4th order finite difference

    c4(1) = 1.37106192
    c4(2) = -0.09322459
    c4(0) = -2*sum(c4(1:2))
	  
  end subroutine

  subroutine init(ds_in, dt_in, nx_in, nz_in)
  ! Initialize grid dimensions, time and spatial steps for the module

    real, intent(in) :: ds_in, dt_in
    integer, intent(in) :: nx_in, nz_in
	
      ds = ds_in
      dt = dt_in
      nx = nx_in
      nz = nz_in
      allocate(vel2(nz,nx), temp(nz,nx))
      temp = 0.
      dt2 = dt*dt

  end subroutine  

  subroutine init_c(c)
    real, intent(in) :: c(0:4)
    c8 = c
  end subroutine

  subroutine one_step8(prev, cur, next)
    real :: prev(nz,nx), cur(nz,nx), next(nz,nx)
    integer :: ix, iz

    do ix=5,nx-4
      do iz=5,nz-4

        next(iz,ix) = vel2(iz,ix)*dt2/ds/ds*(c8(0)*2*cur(iz,ix) &
                        +c8(1)*(cur(iz-1,ix)+cur(iz+1,ix)+cur(iz,ix-1)+cur(iz,ix+1)) &
                        +c8(2)*(cur(iz-2,ix)+cur(iz+2,ix)+cur(iz,ix-2)+cur(iz,ix+2)) &
                        +c8(3)*(cur(iz-3,ix)+cur(iz+3,ix)+cur(iz,ix-3)+cur(iz,ix+3)) &
                        +c8(4)*(cur(iz-4,ix)+cur(iz+4,ix)+cur(iz,ix-4)+cur(iz,ix+4))) &
                      - prev(iz,ix) + 2*cur(iz,ix)
      enddo
    enddo

  end subroutine

  subroutine one_step84(prev, cur, next)
    real, intent(in) :: prev(nz,nx), cur(nz,nx)
    real, intent(inout) :: next(nz,nx)
    integer :: ix, iz

    call Laplacian2(cur,temp)
    call Laplacian2(dt2*vel2*temp,next)

    next = - prev + 2*cur + dt2*vel2*temp + dt2*vel2/12.*next

  end subroutine

  subroutine Laplacian2(cur, Lcur)
    real, intent(in) :: cur(nz,nx)
    real, intent(inout) :: Lcur(nz,nx)
    integer :: ix, iz

    do ix=5,nx-4
      do iz=5,nz-4
        Lcur(iz,ix) = c8(0)*2*cur(iz,ix) &
                        +c8(1)*(cur(iz-1,ix)+cur(iz+1,ix)+cur(iz,ix-1)+cur(iz,ix+1)) &
                        +c8(2)*(cur(iz-2,ix)+cur(iz+2,ix)+cur(iz,ix-2)+cur(iz,ix+2)) &
                        +c8(3)*(cur(iz-3,ix)+cur(iz+3,ix)+cur(iz,ix-3)+cur(iz,ix+3)) &
                        +c8(4)*(cur(iz-4,ix)+cur(iz+4,ix)+cur(iz,ix-4)+cur(iz,ix+4))
      enddo
    enddo
    Lcur = Lcur/ds/ds

  end subroutine

  subroutine clean_up()
    deallocate(temp, vel2)
  end subroutine

end module
